"""
Iterative proportional scaling: fit a precision matrix given its support.

This implements the algorithm 0 in 
  Hara and Takemura, 'A localization approach to improve iterative proportional
  scaling in Gaussian graphical models',
"""

import numpy as np

from sym_inv import sym_inv

def find_cliques(adjacency):
    """
    Search for all maximal cliques in a graph.
 
    This algorithm searches for maximal cliques in a graph.
    maximal cliques are the largest complete subgraph containing
    a given point.  The largest maximal clique is sometimes called
    the maximum clique.
 
    The method essentially unrolls the recursion used in
    the references to avoid issues of recursion stack depth.
    
    Reference::
 
    Based on the algorithm published by Bron & Kerbosch (1973)
        http://doi.acm.org/10.1145/362342.362367
    as adapated by Tomita, Tanaka and Takahashi (2006)
        http://dx.doi.org/10.1016/j.tcs.2006.06.015
    and discussed in Cazals and Karande (2008)
        http://dx.doi.org/10.1016/j.tcs.2008.05.010
 
    This algorithm was borrowed from NetworkX
    """
    # Cache nbrs and find first pivot (highest degree)
    maxconn = -1
    nnbrs = {}
    n_features = adjacency.shape[1]
    adjacency = adjacency.copy()
    adjacency.flat[::n_features+1] = 0
    pivotnbrs = set() # handle empty graph
    for n in range(n_features):
        nbrs = np.where(adjacency[n])[0]
        conn = len(nbrs)
        if conn > maxconn:
            nnbrs[n] = pivotnbrs = set(nbrs)
            maxconn = conn
        else:
            nnbrs[n] = set(nbrs)
    # Initial setup
    cand = set(nnbrs)
    smallcand = cand - pivotnbrs
    done = set()
    stack = []
    clique_so_far=[]
    return_list = list()
    # Start main loop
    while smallcand or stack:
        try:
            # Any nodes left to check?
            n = smallcand.pop()
        except KeyError:
            # back out clique_so_far
            cand, done, smallcand = stack.pop()
            clique_so_far.pop()
            continue
        # Add next node to clique
        clique_so_far.append(n)
        cand.remove(n)
        done.add(n)
        nn=nnbrs[n]
        new_cand = cand & nn
        new_done = done & nn
        # check if we have more to search
        if not new_cand: 
            if not new_done:
                # Found a clique!
                return_list.append(np.array(clique_so_far[:]))
            clique_so_far.pop()
            continue
        # Shortcut--only one node left!
        if not new_done and len(new_cand)==1:
            return_list.append(np.array(clique_so_far + list(new_cand)))
            clique_so_far.pop()
            continue
        # find pivot node (max connected in cand)
        # look in done nodes first
        numb_cand = len(new_cand)
        maxconndone = -1
        for n in new_done:
            cn = new_cand & nnbrs[n]
            conn = len(cn)
            if conn > maxconndone:
                pivotdonenbrs = cn
                maxconndone = conn
                if maxconndone == numb_cand:
                    break
        # Shortcut--this part of tree already searched
        if maxconndone == numb_cand:  
            clique_so_far.pop()
            continue
        # still finding pivot node
        # look in cand nodes second
        maxconn = -1
        for n in new_cand:
            cn = new_cand & nnbrs[n]
            conn = len(cn)
            if conn > maxconn:
                pivotnbrs = cn
                maxconn = conn
                if maxconn == numb_cand-1:
                    break
        # pivot node is max connected in cand from done or cand
        if maxconndone > maxconn:
            pivotnbrs = pivotdonenbrs
        # save search status for later backout
        stack.append( (cand, done, smallcand) )
        cand=new_cand
        done=new_done
        smallcand = cand - pivotnbrs

    return return_list


def prop_scaling(cov, support, max_iter=100, eps=1e-12, prec_init=None, 
                cliques=None):
    """ Iterative proportional scaling: given a covariance and a support 
        find the closest precision matrix with the given support.
    """
    n_features = cov.shape[0]
    if np.all(support != 0):
        return sym_inv(cov)
    if cliques is None:
        cliques = find_cliques(support)
    cliques.sort(key=len, reverse=True)
    all_indices = set(range(n_features))
    if prec_init is None:
        # Heuristic to get a good initialization
        prec = support.astype(np.float)/n_features
        prec.flat[::n_features+1] = 1
    else:
        prec = prec_init.copy()
    # A dictionnary to cache inverse empirical covariance subsect
    inv_covs = dict()
    for i in range(max_iter):
        old_prec = prec.copy()
        for clique in cliques:
            # A hashable marker for the clique
            clique_set = frozenset(clique)
            not_clique = all_indices.difference(clique)
            not_clique = np.array(list(not_clique))
            subset = (clique[:, np.newaxis], clique)
            if clique_set not in inv_covs:
                inv_cov = sym_inv(cov[subset]) 
                inv_covs[clique_set] = inv_cov
            else:
                inv_cov = inv_covs[clique_set]
            new_prec = prec.copy()
            sub_prec = prec[clique[:, np.newaxis], not_clique]
            new_prec[subset] = (inv_cov
                     + np.dot(np.dot(sub_prec,
                        sym_inv(prec[not_clique[:, np.newaxis], not_clique])),
                            sub_prec.T,
                     ))
            this_n_features = len(clique)
            if ((prec - new_prec)**2).sum() < .1*eps*this_n_features**2 and i > 2:
                prec = .5*(new_prec + new_prec.T)
                break
            prec = .5*(new_prec + new_prec.T)
        if ((prec - old_prec)**2).sum() < eps*n_features**2:
            break
        # XXX: probably need to compute the dual gap

    return prec


