"""
Basic statistical tools for Gaussian multivariate models.
"""

import numpy as np
from scipy import linalg, signal, fftpack

def rescale_mat(mat, scale_inv=False):
    """ Rescale a matrix by putting its diagonal to 1.
    """
    if scale_inv:
        mat = linalg.inv(mat)
    diag = np.diag(mat)
    d = np.sign(diag)/np.sqrt(np.abs(diag))
    mat = (mat*d).T*d
    if scale_inv:
        mat = linalg.inv(mat)
    return mat


################################################################################
# Graph likelihoods

def log_likelihood(data, precision, lambdas=None, data_is_cov=False,
                    use_correlations=False):
    """ Evaluates the log-likelihood of the data using the given
        precision matrix. The value is returned compared to a
        identity-precision null hypothesis.

        Parameters
        ----------
        data: 2D ndarray
            The multivariate samples to evaluate log-likelihood upon.
        precision: 2D ndarray
            The precision matrix giving the graph model
        lambdas: 1D ndarray, optional
            The eigenvalues of the precision matrix.
        data_is_cov: boolean, optional
            If True, the data argument is conisdered as the empirical 
            covariance matrix of the data matrix.
        use_correlations: boolean, optional
            If True, the given precision matrix and the data are scaled
            to compare correlations, rather than covariance 
            (diag(cov) = 1)
    """
    n_features, n_samples = data.shape
    if data_is_cov:
        assert n_features == n_samples, ('data_is_cov is True, but the '
                'data passed is not square')
        emp_cov = data
    else:
        emp_cov = np.dot(data, data.T)/n_samples
    if np.any(np.logical_not(np.isfinite(emp_cov))):
        return -np.inf
    if use_correlations:
        d = 1/np.sqrt(np.diag(emp_cov))
        if np.any(np.logical_not(np.isfinite(d))):
            return -np.inf
        emp_cov = (emp_cov*d).T*d
        model_cov = linalg.inv(precision)
        d = 1/np.sqrt(np.diag(model_cov))
        if np.any(np.logical_not(np.isfinite(d))):
            return -np.inf
        model_cov = (model_cov*d).T*d
        precision = linalg.inv(model_cov)
    if lambdas is None:
        # Use eigenvalues to calculation log-determinant, because 
        # elsewhere it overflows.
        lambdas = linalg.eigvals(precision).real
        if np.any(lambdas <= 0):
            return -np.inf
    return 0.5*(
            np.log(lambdas).sum() +
            n_features -
            np.trace(np.dot(emp_cov, precision))
        )


def kl_divergence(estimated_precision, reference_precision):
    """ Evaluates the Kullback-Leibler between two Gaussian graphical
        models described by the precisions.
    """
    n_features, _ = estimated_precision.shape
    # Use eigenvalues to calculation log-determinant, because 
    # elsewhere it overflows.
    return 0.5*(
              np.log(linalg.eigvals(estimated_precision).real).sum()
            - np.log(linalg.eigvals(reference_precision).real).sum()
            - n_features
            + np.trace(np.dot(np.linalg.inv(estimated_precision),
                            reference_precision))
        )


################################################################################
# Covariance estimators



def ledoit_wolf(x, return_factor=False):
    """ Estimates the shrunk Ledoit-Wolf covariance matrix.

        Parameters
        ----------
        x: 2D ndarray, shape (p, n)
            The data matrix, with p features and n samples.
        return_factor: boolean, optional
            If return_factor is True, the regularisation_factor is
            returned.

        Returns
        -------
        regularised_cov: 2D ndarray
            Regularized covariance
        regularisation_factor: float
            Regularisation factor

        Notes
        -----
        The regularised covariance is::

            (1 - regularisation_factor)*cov 
                    + regularisation_factor*np.identity(n_features)
    """
    n_features, n_samples = x.shape
    if n_features == 1:
        if return_factor:
            return np.atleast_2d(x.std()), 0
        return np.atleast_2d(x.std())
    cov = np.dot(x, x.T)/n_samples
    i = np.identity(n_features)
    mu = np.trace(cov)/n_features
    delta = ((cov - mu*i)**2).sum()/n_features
    #beta_ = 1./(n_features*n_samples**2) * sum([
    #        ((np.dot(this_x[:, np.newaxis], this_x[np.newaxis, :]) - cov)**2).sum()
    #        for this_x in x.T
    #    ])
    x2 = x**2
    beta_ = 1./(n_features*n_samples) * np.sum(
                            np.dot(x2, x2.T)/n_samples - cov**2
                )

    beta = min(beta_, delta)
    alpha = delta - beta
    if not return_factor:
        return beta/delta * mu * i + alpha/delta * cov
    else:
        return beta/delta * mu * i + alpha/delta * cov, beta/delta


def shrunk_covariance(x, shrinkage=0.1, use_correlations=False,
                        data_is_cov=False):
    """ Calculate a covariance matrix shrunk on the diagonal.
    """
    n_features, n_samples = x.shape
    if data_is_cov:
        emp_cov = x
    else:
        emp_cov = np.dot(x, x.T)/n_samples
    if use_correlations:
        d = 1/np.sqrt(np.diag(emp_cov))
        emp_cov = (emp_cov*d).T*d
    shrunk_cov = ((1-shrinkage)*emp_cov + shrinkage*np.eye(n_features))
    return shrunk_cov

################################################################################
# Estimators from time series

def standardize(signals, copy=True, normalize=True):
    """ Center and norm a given signal (sample = axis -1)
    """
    # FIXME: Should this be merged with canica.algorithms.center_and_norm
    signals = np.array(signals, copy=copy)
    length = float(signals.shape[-1])
    buffer  = signals.T
    buffer -= signals.mean(axis=-1)
    if normalize:
        std = np.sqrt(length)*signals.std(axis=-1).T
        std[std==0] = 1
        buffer /= std
    return signals


def clean_signals(signals, confounds=None, low_pass=0.2, t_r=2.5,
                         high_pass=False, detrend=False, 
                        normalize=True,
                        shift_confounds=False):
    """ Normalize the signal, and if any confounds are given, project in
        the orthogonal space.

        Low pass filter improves specificity (more interesting arrows 
        selected)

        High pass filter should be kepts small, so as not to kill
        sensitivity
    """
    signals = standardize(signals, normalize=normalize)

    if confounds is not None:
        # Restrict the signal to the orthogonal of the confounds
        confounds = np.atleast_2d(confounds)
        if shift_confounds:
            confounds = np.r_[confounds[..., 1:-1],
                              confounds[..., 2:],
                              confounds[..., :-2]]
            signals = signals[..., 1:-1]
        confounds = standardize(confounds, normalize=True)
        confounds = linalg.qr(confounds.T, econ=True)[0].T
        signals -= np.dot(np.dot(signals, confounds.T), confounds)
    
    if low_pass or high_pass:
        n = signals.shape[-1]
        freq = fftpack.fftfreq(n, d=t_r)
        for s in signals:
            fft = fftpack.fft(s)
            if low_pass:
                fft[np.abs(freq) > low_pass] = 0
            if high_pass:
                fft[np.abs(freq) < high_pass] = 0
            s[:] = fftpack.ifft(fft)

    if detrend:
        for s in signals:
            s[:] = signal.detrend(s)

    signals = standardize(signals, normalize=normalize)
    return signals


def corrcoef(data, y=None, rowvar=1, ld=True, bias=0, cov=False):
    """ Return the correlation matrix, optionally using the Ledoit-Wolf
        estimator.

        Parameters
        ----------
        data: 1D or 2D array
            array: shape: (feature, samples)
        y : array_like, optional
            An additional set of variables and observations. `y` has the 
            same form as that of `data`.
        ld: boolean, optional
            if ld, use the Ledoit-Wolf regularisation.
        rowvar : int, optional
            If `rowvar` is non-zero (default), then each row represents a
            variable, with observations in the columns. Otherwise, the 
            relationship is transposed: each column represents a variable, 
            while the rows contain observations.
        bias : int, optional
            Default normalization is by ``(N-1)``, where ``N`` is the number 
            of observations given (unbiased estimate). If `bias` is 1, 
            then normalization is by ``N``.
        cov : boolean, optional
            If cov is True, the covariance is calculated, rather than the 
            correlation.

        Returns
        -------
        out : ndarray
            The covariance matrix of the variables.
        
        Notes
        ------
        This function is the same as numpy.cov, appart from the fact that 
        it uses the Ledoit-Wolf estimator.
    """
    if not ld:
        return np.corrcoef(data, y=y, rowvar=rowvar, bias=bias)
    data = np.array(data, copy=True, ndmin=2, dtype=np.float)
    if data.shape[0] == 1:
        rowvar = 1
    if rowvar:
        axis = 0
    else:
        axis = 1
    if y is not None:
        y = np.array(data, copy=False, ndmin=2, dtype=np.float)
        data = np.concatenate((data, y), axis)
    if not rowvar:
        data = data.T
    if not cov:
        data -= data.mean(axis=-1)[..., np.newaxis]
        std = data.std(axis=-1)[..., np.newaxis]
        # Cull out 0 std values.
        std[std==0] = 1
        data /= std
    cor = ledoit_wolf(data)
    return cor


